#ifndef RLIST_HH
#define RLIST_HH
//
//    This is a root-compatible version of a subset of the standard 
//    template library list class.
//
//----------------------------------  List node template
namespace std {

template<class T>
class list_node {
public:
  list_node(const T& x)
    : mNext(0), mPrev(0), mData(x)
    {}

  ~list_node() {
      if (mNext) mNext->mPrev = mPrev;
      if (mPrev) mPrev->mNext = mNext;
  }

  list_node* getNext() const {return mNext;}
  list_node* getPrev() const {return mPrev;}
  T& getData() {return mData;}
  const T& getData() const {return mData;}
  list_node* after(const T& x) {
      list_node* r = new list_node(x);
      if (mNext) mNext->mPrev = r;
      r->mNext = mNext;
      r->mPrev = this;
      mNext = r;
      return r;
  }
  list_node* before(const T& x) {
      list_node* r = new list_node(x);
      if (mPrev) mPrev->mNext = r;
      r->mNext = this;
      r->mPrev = mPrev;
      mPrev = r;
      return r;
  }
private:
  list_node* mNext;
  list_node* mPrev;
  T mData;
};

//-----------------------------------  List class template.
template<class T> 
class list_pointer {
public:
  typedef list_node<T> node_t;

  list_pointer() : mPointer(0) {}

  list_pointer(const list_pointer& p) {mPointer = p.mPointer;}

  list_pointer(node_t* p) {mPointer = p;}

  void remove() {delete mPointer;}

  list_pointer operator=(const list_pointer& p) { // comparison !=
      mPointer = p.mPointer;
      return *this;}

  bool operator!=(const list_pointer& p) const {  // comparison !=
      return (mPointer != p.mPointer);}

  bool operator==(const list_pointer& p) const {  // comparison ==
      return (mPointer == p.mPointer);}

  list_pointer operator++(void) {                 // pre-increment
      return (mPointer = mPointer->getNext());}

  list_pointer operator++(int)  {                 // post-increment
      node_t* save = mPointer;
      mPointer = mPointer->getNext();
      return save;}

  T& operator*(void) const {return mPointer->getData();}

  T* operator->(void) const {return &mPointer->getData();}

  list_pointer next(void) const {
      if (mPointer) return mPointer->getNext();
      return 0;
  }

  node_t* getNode() const {return mPointer;}

private:
  node_t* mPointer;
};

//-----------------------------------  List class template.
template<class T> 
class const_list_pointer {
public:
  typedef list_node<T> node_t;

  const_list_pointer() {mPointer = 0;}

  const_list_pointer(const list_pointer<T>& p) {mPointer = p.getNode();}

  const_list_pointer(const node_t* p) {mPointer = p;}

  bool operator!=(const const_list_pointer& p) const {  // comparison !=
      return (mPointer != p.mPointer);}

  bool operator==(const const_list_pointer& p) const {  // comparison ==
      return (mPointer == p.mPointer);}

  const_list_pointer operator++(void) {                 // pre-increment
      return (mPointer = mPointer->getNext());}

  const_list_pointer operator++(int)  {                 // post-increment
      const node_t* save = mPointer;
      mPointer = mPointer->getNext();
      return save;}

  const T& operator*(void) const {return mPointer->getData();}

  const T* operator->(void) const {return &mPointer->getData();}

  const const_list_pointer next(void) const {
      if (mPointer) return mPointer->getNext();
      return 0;
  }

  const node_t* getNode() const {return mPointer;}

private:
  const node_t* mPointer;
};

//-----------------------------------  List class template.
template<class T>
class list {
public:
  typedef list_pointer<T> iterator;
  typedef const_list_pointer<T> const_iterator;
  typedef list_node<T>    node_type;
  list() {mNode = 0;}
  ~list() {};

  void push_back(const T& x) {
      if (mNode != 0) last().getNode()->after(x);
      else            mNode = new node_type(x);
  }

  void push_front(const T& x) {
      if (mNode == end()) mNode = new node_type(x);
      else                mNode = iterator(mNode.getNode()->before(x));
  }

  void pop_front(void) {
      erase(begin());
  }

  iterator begin() {return mNode;}
  const iterator begin() const {return mNode;}

  iterator end() {return iterator(0);}
  const iterator end() const {return iterator(0);}

  void erase(iterator node) {
       if (node == mNode) mNode++;
       node.remove();
  }

  void erase(iterator f, iterator l) {
       for (iterator x=f ; x!=l && x!=end() ;) {
	   iterator save = x++;
	   this->erase(save);}
  }

  void insert(iterator i, const T& x) {
      if (i == end())        push_back(x);
      else if (i == begin()) push_front(x);
      else                   i.getNode()->before(x);
  }

  bool empty(void) const {return (begin() == end());}

  const T& front(void) const {return *const_iterator(mNode);}

  T& front(void) {return *mNode;}

  T& back(void) {return *last();}
  const T& back(void) const {return *last();}

  const_iterator last(void) const {
      const_iterator p(mNode);
      for ( ; p.next() != end() ; p++);
      return p;
  }

  iterator last(void) {
      iterator p(mNode);
      for ( ; p.next() != end() ; p++);
      return p;
  }

private:
  iterator mNode;
};

}
#endif  /// RLIST_HH
