#include "PTmon.hh"
#include "PTChan.hh"

#ifndef __CINT__

//-- The next three lines are needed for trigger generation

#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/PTmon/PTmon.cc,v 1.18 2005/10/01 01:12:20 nzotov Exp $"
#define PIDTITLE "Peak-Trough glitch monitor"
#include "ProcIdent.hh"
#include "TrigBase.hh"
#include "TrigRslt.hh"

#include <string.h>
#include <stdlib.h>
#include <cmath>
#include <list>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <vector>

#include "Dacc.hh"
#include "ParseLine.hh"
#include "TSeries.hh" 

//--------------------------------------- Include C headers 
 
#include "detect.h"
#include "ligoconfig.h"
#include "ligocont.h"
#include "ligostd.h"

//--------------------------------------- Generate the main routine
EXECDAT(PTmon)
#endif                                // !def(__CINT__)
                                       
using namespace std;

//--------------------------------------- Skeleton object constructor
PTmon::PTmon(int argc, const char *argv[])
  : DatEnv(argc, argv), mStride(1.0), mOSC(getDacc())
{
 
  //------------------------------------- Look for arguments
    const char* cfile = "PTmon.conf";
    bool syntax = false;
    for (int i=1; i<argc; i++){
        if (!strcmp("-conf", (argv[i]))) {
            cfile = argv[++i];
        } else if (isDatEnvArg(argv[i])) {
            i++;
        } else {
            syntax = true;
            cerr << "Error in command.  Syntax:" << endl;
            cerr << argv[0] << " [-conf <cfiles>] " << endl;
            finish();
            return;
        }
    }
    mCFile = cfile;
    MonServer::setDebug(Debug());

//-------------------------------------- Define and initialize parameters
    mDict.addPar("Stride",     double(1.0));
    mDict.addPar("TrigEnable", 0);
    mDict.addPar("MonName",    string(argv[0]));
    mDict.addPar("OSCFile",    string(""));

//-------------------------------------- Read configuration file and set up
    if (ReadConfig()) {
        finish();
        return;
    }

//-------------------------------------- Set time stride and alignment
    getDacc().setStride(mStride);       // before calling ProcessData
    setStrideAlignment(1,0);

//-------------------------------------- Set up the rest
    if (!mDict.getString("OSCFile").empty()) {
      mOSC.readConfig(mDict.getString("OSCFile").c_str());
    }
    getDacc().setIgnoreMissingChannel(true);
}
//========================================== Skeleton object destructor
PTmon::~PTmon(){
    Reset();
}

//======================================= Interrupt handler
void
PTmon::Attention(void) {
    if (Debug()) cout << "PTmon: Attention Interrupt" << endl;
    MonServer::Attention();
}

//======================================== Read in configuration file
bool
PTmon::ReadConfig(void) {
    Reset();
 
//---------------------------------------- Set up line parser
    ParseLine lp(mCFile.c_str());
    if (!lp.isOpen()) {
        cerr << "Unable to open configuration file" << mCFile << endl;
        return true;
    }   
    ofstream* confout=0;

    if (Debug()) {
        lp.setLog(cout); 
    }
    for (int nArg = lp.getLine(); nArg>=0; nArg = lp.getLine()) {

//----------------------------------- Empty command line 
        if (!lp[0]) {
            continue;
	
//----------------------------------- Parse parameter line
        } else if (!strcmp(lp[0], "Parameter")) {
	    for (int i=1; i<nArg; i+=2) {
	        if (!mDict.exists(lp[i])) {
                    cout << "No such parameter: " << lp[i] << endl;
                    break;
                } else {
                    if (Debug()) cout << "Set parameter: " << lp[i] << endl;
                    mDict.setPar(lp[i], lp[i+1]);
                }
            }   
	
//----------------------------------- Parse channel definition
        } else if (!strcmp(lp[0], "Channel")) {
	    int inx = 3;

	    if (nArg < 2) {
	        cout << "Channel not specified" << endl;
	        continue;
	    } else if (nArg == 2 || *lp[2] == '-') {
                mChannels[string(lp[1])] = PTChan(lp[1]);
                inx = 2;
            } else {
	          mChannels[string(lp[1])] = PTChan(lp[2]);                   
            }

            bool syntax = false;
            char *p, *q, *r;
	    for (int i=inx; i<nArg; i++) {
                string lpi = lp[i];
                if (lpi == "-sps") {
                    long int samSec = (int)strtod(lp[++i], &p);
                    syntax |= *p;
                    mChannels[lp[1]].setSps(samSec);
		} else if (lpi == "-threshold") {
		    float repThresh = (float)strtod(lp[++i], &q);
		    syntax |= *q;
		    mChannels[lp[1]].setThreshold(repThresh);
		} else if (lpi == "-settle"){
		    dt = (double)strtod(lp[++i], &r);
		    syntax |= *r;
		    mChannels[lp[1]].setSettle(dt);
		} else if (lpi == "-trigger") {
 		    mChannels[lp[1]].setTrigger(true);
		} else if (lpi == "-while") {
		    mChannels[lp[1]].setEnableCond(lp[++i]);
		}else{
                    syntax = true;
                    cout << "Invalid channel argument: " << endl;
		}   
		if (syntax) {
		    cout << "Channel syntax is:" << endl;
		    cout << "Channel <chan> -sps <sampling rate> ";
		    cout << "-threshold <threshold>  -settle <filter settling time> ";
		    cout << " -trigger" << endl;
                            
 	        }
	     }
          }      
     }       

//----------------------------------- Close out configuration output
      if (confout) {
          confout->close();
          delete confout;
      }

//---------------------------------------- Set the time stride
    getDacc().setStride(Interval(mDict.getDouble("Stride")));

//-------------------------------------- Connect channels to Dacc

     for (chan_iter i = mChannels.begin(); i != mChannels.end(); i++){
          i->second.setup(getDacc());
     }
     return false; 
}

//======================================= Clear all lists
void
PTmon::Reset(void){ 

//--------------------------------------- Delete Dacc channels
    while (mChannels.begin() !=mChannels.end()) {
        mChannels.erase(mChannels.begin()); 
    }
}
//======================================  Frame processing function.
void 
PTmon::ProcessData(void) {
 
 const int maxOnset = 20;

//-------------------------------------- Connect channels to Dacc & process

  for (chan_iter i = mChannels.begin(); i != mChannels.end(); i++){
      i->second.reset();
      const char* condName = i->second.getEnableCond();

      if(!*condName || mOSC.satisfied(condName)) {
        s_onset *oPtr = new s_onset[maxOnset]; // Set up vector for events

        if (i->second.findit(maxOnset, oPtr)) {
             i->second.refOnsetData();


//--------------------------------------- Process glitches -> triggers

          if (mDict.getInt("TrigEnable") && i->second.getTrigger()) {
              const char* chan = (i->second.getChannel());
	      trig::TrigRslt t("PTmon",chan,0,0);

    	      if (chan && chan[2] == ':') {
	          string ifo(chan, 2);
	          t.setIfos(ifo.c_str());
	      }

	      for (int j=0; j< i->second.mOnsetN; j++){
                if (oPtr[j].o_maxamp >= i->second.mThresh){ 
	          t.setTime(Time(oPtr[j].o_secs, oPtr[j].o_ns));	     
                  t.setPriority(trig::p_error);
	          t.setDisposition(trig::d_metaDB);
	          t.setIntensity(oPtr[j].o_maxamp);
     	          t.setFrequency(oPtr[j].o_freq);
    	          int rc = sendTrigger(t);
       	          if (rc) {                             
                    cout << "Error " << rc << " logging PTmon trigger" << endl;
	          }
		}
              }
          }     // Trigger writing completed   
      }         // Back from 'findit' (glitch found)
      delete [] oPtr; 
     }          // Check OSC 
   }            // Loop through channels 
}               // End of ProcessData

//======================================== Test if channel exists
bool
PTmon::chanExists(const string& key) const {
cout << "In chanExists" << endl;
    return mChannels.find(key) != mChannels.end();
}
