//
//    File Slice2.cc
//
//    Data slice monitor: Look for repeated 1/16th second slices of data.
//
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include "Dacc.hh"
#include "ParseLine.hh"
#include "Slice2.hh"
#include "Time.hh"
#include "TrigRslt.hh"
#include "framecpp/FrameH.hh"

//------------------------------------  Generate the main function.
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/Slice2/Slice2.cc,v 1.14 2008/07/18 17:15:29 jzweizig Exp $"
#define PIDTITLE  "Watch for misplaced slices"
#include "ProcIdent.hh"

using namespace std;
using FrameCPP::FrameH;
using FrameCPP::FrAdcData;

EXECDMT(Slice2)

//------------------------------------  Constructor
Slice2::Slice2(int argc, const char *argv[]) 
 : DMTBase(argc, argv), mGenTrig(false), mAlign(true), mSkipMin(2), 
   mTrigMask(~0), mGroupSecs(1.0), mFrameCt(0), mSumTimes(0), mLastTime(0),
   mTrigCt(0), mLastTrig(0)
{
    //--------------------------------  Parse the arguments
    mConfig = "Slice2.conf";
    for (int i=1 ; i<argc ; i++) {
        string argi(argv[i]);
        if (argi == "cfile") {
	    mConfig = argv[++i];
	} else if (argi.substr(0,5) == "+trig") {
	    mGenTrig = true;
	    if (*(argv[i]+5) == ':') mTrigMask = strtol(argv[i]+6, 0, 0);
	} else if (argi == "group") {
	    mGroupSecs = strtod(argv[++i],0);
	} else if (argi == "nskip") {
	    mSkipMin = strtol(argv[++i],0,0);
	} else if (isEnvArg(argv[i])) {
	    i++;
	} else {
	    cerr << "Undefined argument: " << argi << endl;
	    cerr << "Syntax for " << argv[0] << ":" << endl;
	    cerr << argv[0] << " [cfile <conf>] [+trig] " 
		 << "[nskip <min-skips>] [<DMTBase-args>]" << endl;
	    finish();
	    return;
	}
    }

    //----------------------------------  Read the configuration file
    if (ReadConfig()) {
        finish();
	return;
    }
}

//======================================  Read in the configuration
bool
Slice2::ReadConfig(void) {
    if (Debug()) {
        cout << "Slice2: Reading configuration from " << mConfig << endl;
    }
    ParseLine p(mConfig.c_str());
    if (Debug()) p.setLog(cout);
    for (int nArg=p.getLine() ; nArg >= 0 ; nArg=p.getLine()) {
        Interval tSlice(0.0625);
	Interval tMaxOff(0.25);
        if (!nArg) continue;
	if (nArg > 1) tSlice  = Interval(strtod(p[1], 0));
	if (nArg > 2) tMaxOff = Interval(strtod(p[2], 0));
	if (nArg > 3) {
	    cerr << "Unexpected argument" << endl;
	    return true;
	}
	ChanSlice sl(p[0]);
	if (!tSlice || !tMaxOff) {
	    cerr << "Invalid time specified, must be > 0" << endl;
	    return true;
	}
	sl.setSliceTime(tSlice, tMaxOff);
	sl.setDebug(Debug());
        mChannel.push_back(sl);
    }
    return false;
}

//======================================  Process one frame (FrameCPP).
void
Slice2::ProcessFrame(const FrameH* frame) {

    //----------------------------------  Check that the start time is aligned
    Time tStart(frame->GetGTime().getSec(), frame->GetGTime().getNSec());
    if (tStart.getS() == 0) {
        if (Debug()) cout << "Slice2: Frame time is zero" << endl;
        if (trigEnable(0x40) && mAlign) {
	    trig::TrigRslt t("FrameTimeError", "TimeIsZero");
	    t.setTime(Now());
	    t.setPriority(trig::p_error);
	    t.setDisposition(trig::d_metaDB |trig::d_alarm);
	    t.setDuration(0.0);
	    sendTrigger(t);
	    mAlign = false;
	}
    } else if (tStart.getN() != 0) {
        if (Debug()) cout << "Slice2: Frame time offset is: " << tStart.getN()
			  << " ns."<< endl;
        if (trigEnable(0x20) && mAlign) {
	    trig::TrigRslt t("FrameTimeError", "Misaligned");
	    t.setTime(tStart);
	    t.setPriority(trig::p_error);
	    t.setDisposition(trig::d_metaDB |trig::d_alarm);
	    t.setDuration(0.0);
	    sendTrigger(t);
	    mAlign = false;
	}
    } else {
        mAlign = true;
    }

    //----------------------------------  Check the frame duration is 1 second
    Interval dT(frame->GetDt());
    int nGroups = int(dT / mGroupSecs);
    if (dT > mGroupSecs*double(nGroups)) nGroups++;
    std::vector<int> grpCount(nGroups, 0);

    //if (dT != Interval(1.0)) {
    //    if (Debug()) cout << "Slice2: Frame duration is " 
    //		  << double(dT) << endl;
    //    if (trigEnable(0x10)) {
    //    trig::TrigRslt t("FrameTimeError", "Not_1_second");
    //    t.setTime(tStart);
    //    t.setPriority(trig::p_error);
    //    t.setDisposition(trig::d_metaDB | trig::d_alarm);
    //    t.setDuration(dT);
    //    sendTrigger(t);
    //}
    //}

    //----------------------------------  Loop over channels
    int nBadChan(0);
    for (channel_iter itr=mChannel.begin() ; itr!=mChannel.end() ; itr++) {
        const char* ChName = itr->getChannel();
	channel_iter tent = itr;

        //------------------------------  Get the adc pointer.
	const FrAdcData* adc = getDacc().nextAdc(ChName);
	if (!adc) {
	    adc = getDacc().findAdc(ChName);
	    if (!adc) {
	        if (!mFrameCt) cout << "Missing channel " << ChName << endl;
		continue;
	    } else if (itr != mChannel.begin()) {
	        channel_iter save = tent;
	        --itr;
		tent = mChannel.insert(itr, *save);
		mChannel.erase(save);
	    }
	}

	if (adc->GetDataValid()) {
	    if (Debug()) {
	        cout << "Channel: " << ChName << " Readout status: " 
		     << adc->GetDataValid() << endl;
	    }
	    tent->incError();
	    if (!nBadChan++ && trigEnable(0x08)) {
	        trig::TrigRslt t("ChannelReadOutError", ChName);
		t.setTime(tStart);
		t.setPriority(trig::p_error);
		t.setDisposition(trig::d_metaDB | trig::d_alarm);
		t.setDuration(dT);
		sendTrigger(t);
	    }
	    continue;
	}

	//------------------------------  Test that offset is inside frame
	Interval tOff(adc->GetTimeOffset());
	if ((tOff < Interval(0.0) || tOff > dT) && trigEnable(0x04)) {
	    trig::TrigRslt t("TimeSkewError", ChName);
	    t.setTime(tStart);
	    t.setPriority(trig::p_error);
	    t.setDisposition(trig::d_metaDB | trig::d_alarm);
	    t.setDuration(dT);
	    sendTrigger(t);
	}

	if (adc->RefData().size() == 0) continue;
        const FrameCPP::FrVect* vect = adc->RefData().front();
	if (!vect) continue;

	double sampRate = adc->GetSampleRate();
	tent->setSampleRate(sampRate);

	const_cast<FrameCPP::FrVect*>(vect)->Uncompress();
	int nw = vect->GetNData();
	if (nw <= 0) continue;

	//------------------------------  Test that time span is inside frame
	if (Interval(double(nw)/sampRate) + tOff > dT) {
	    if (Debug()) cout << "Slice2: Error in channel " << ChName 
			      << " time interval: nw=" << nw 
			      << " Rate=" << sampRate << " Offset=" 
			      << tOff << " dT=" << dT << endl;
	    if (trigEnable(0x02)) {
	        trig::TrigRslt t("TimeIntervalError", ChName);
		t.setTime(tStart);
		t.setPriority(trig::p_error);
		t.setDisposition(trig::d_metaDB | trig::d_alarm);
		t.setDuration(dT);
		sendTrigger(t);
	    }
	}

        const void* data =vect->GetData();
	if (!data) continue;

	//------------------------------  Process depending on type
	int vtype = vect->GetType();
	int nWGroup = int(sampRate * mGroupSecs);
	for (int iGrp=0; iGrp < nGroups; iGrp++) {
	    int nSkip(0);
	    int iOff = nWGroup * iGrp;
	    int iw   = nWGroup;
            if (iOff + iw > nw) iw = nw - iOff;
	    switch (vtype) {
	    case FrameCPP::FrVect::FR_VECT_2S:
	    case FrameCPP::FrVect::FR_VECT_2U:
	        nSkip=tent->Scan(iw,reinterpret_cast<const short*>(data)+iOff);
		break;
	    case FrameCPP::FrVect::FR_VECT_4R:
	        nSkip=tent->Scan(iw,reinterpret_cast<const float*>(data)+iOff);
		break;
	    case FrameCPP::FrVect::FR_VECT_8C:
	        nSkip=tent->Scan(iw,reinterpret_cast<const fComplex*>(data)+iOff);
		break;
	    default:
	        cerr << "Slice2: Unable to process data from " << ChName 
		     << endl;
		break;
	    }
	    if (nSkip) grpCount[iGrp]++;
	}
    }

    //-----------------------------------  Generate a trigger if the error    
    for (int iGrp=0; iGrp < nGroups; iGrp++) {
        if (grpCount[iGrp] >= mSkipMin) {
	    Time t0 = tStart + mGroupSecs*double(iGrp);
	    if (trigEnable(0x01)) {
	        trig::TrigRslt t("TimeSliceError", "MultiChannel");
		t.setTime(t0);
		t.setDuration(mGroupSecs);
		t.setPriority(trig::p_error);
		t.setDisposition(trig::d_metaDB | trig::d_alarm);
		sendTrigger(t);
	    } else {
	        cout << "Time slice errors found in " << grpCount[iGrp]
		     << " channels at " << t0 << endl;
	    }

	    //----------------------------  Count triggers
	    mTrigCt++;
	    mLastTrig++;
	    for (channel_iter i=mChannel.begin() ; i!=mChannel.end() ; i++) {
	        i->countTrigger();
	    }
	}
    }

    //----------------------------------  Generate a report every 300s
    mFrameCt++;
    mSumTimes += dT;
    mLastTime += dT;
    if (!(mFrameCt%300)) Report();
}

//--------------------------------------  Attention handler.
void 
Slice2::Attention(void) {
    Report();
    mChannel.erase(mChannel.begin(), mChannel.end());
    ReadConfig();
}

//======================================  Write a status report
void
Slice2::Report(void) {
    char line[128];

    //----------------------------------  Generate a file name
    if (getenv("DMTHTMLOUT")) {
        sprintf(line, "%s/Slice2.report",getenv("DMTHTMLOUT"));
    } else {
        sprintf(line, "Slice2.report");
    }

    //----------------------------------  Open the output stream
    ofstream out;
    out.open(line);
    if (out.bad()) {
        cerr << "Unable to open report file: " << line << endl;
	return;
    }

    //----------------------------------  Collect global stats
    double sumSingle = 0.0;
    double sumSquare = 0.0;
    double sumError  = 0.0;
    for (channel_iter itr=mChannel.begin() ; itr!=mChannel.end() ; itr++) {
	double rate = itr->getSingleRate();
	sumSingle += rate;
	sumSquare += rate*rate;
	sumError  += itr->getErrorTot();
    }

    //----------------------------------  Print Slice2 parameters, etc.
    LocalStr(Now(), line, "%M %d, %Y %H:%N");
    out << "    Slice2 Report: " << line      << endl << endl;
    out << "Slice time:            " << 0.0625   << "s" << endl;
    out << "Maximum time shift:    " << 0.25     << "s" << endl;
    out << "Total frames read:     " << mFrameCt << endl;
    out << "Total time processed:  " << mSumTimes << endl;
    out << "R/O errors per frame:  " << sumError/mFrameCt  << endl;
    out << "Total triggers:        " << mTrigCt  << endl;
    out << "Total rate:            " << mTrigCt/double(mSumTimes) << endl;
    out << "Triggers in last 300s: " << mLastTrig << endl;
    out << "Rate in last 300s:     " << mLastTrig/double(mLastTime) << endl;
    out << "Trigger multiplicity:  " << mSkipMin << endl;
    out << "Sum singles rate:      " << sumSingle << endl;
    double unbiased = 0.5 * (sumSingle*sumSingle - sumSquare);
    out << "Uncorrelated doubles:  " << unbiased << endl;
    out << endl;

    //----------------------------------  Print channel summaries.
    out << "Channel                         Frames    Singles   Triggers     Errs/f"
	<< endl;
    for (channel_iter itr=mChannel.begin() ; itr!=mChannel.end() ; itr++) {
	sprintf(line, "%-31s %6i %10f %10f %10f", 
		itr->getChannel(), itr->getNTest(), 
		itr->getSingleRate(), itr->getTriggerRate(), 
		itr->getErrorRate());
	out << line << endl;
	itr->resetStats();
    }
    mLastTrig = 0;
    mLastTime = 0.0;
    out.close();
}

//--------------------------------------  Print final Results
Slice2::~Slice2()
{
    Report();
    mChannel.clear();
}

//------------------------------------  Channel constructor
ChanSlice::ChanSlice(const char *name) 
  : mChanName(name), mSaved(0), mSavType(kNone), mSavData(0), mRate(0.0), 
    mSliceTime(0.0), mNWSlice(0), mNCkSlice(0), mDebug(0)
{
}

ChanSlice::ChanSlice(const ChanSlice& x) 
  : mSavType(kNone)
{
    *this = x;
}


ChanSlice&
ChanSlice::operator=(const ChanSlice& x) {
    mChanName  = x.mChanName;
    mRate      = x.mRate;
    mSliceTime = x.mSliceTime;
    mNWSlice   = x.mNWSlice;
    mNCkSlice  = x.mNCkSlice; 
    mDebug     = x.mDebug;
    mNOverlap  = x.mNOverlap;
    mStats.clear();
    mTotal.clear();
    mStats += x.mStats;
    mTotal += x.mTotal;
   
    switch (x.mSavType) {
    case kNone:
        freeData();
	break;
    case kShort:
        saveData(reinterpret_cast<const short*>(mSavData), mSaved);
	break;
    case kFloat:
        saveData(reinterpret_cast<const float*>(mSavData), mSaved);
	break;
    case kCmplx:
        saveData(reinterpret_cast<const fComplex*>(mSavData), mSaved);
	break;
    }
    return *this;
}

//------------------------------------  Channel destructor
ChanSlice::~ChanSlice() {
    freeData();
}


//------------------------------------  Set sample rate
void
ChanSlice::setSampleRate(double Rate) {
    mRate = Rate;
    if (mSliceTime > Interval(0.0)) {
        mNWSlice=size_t(mRate*double(mSliceTime) + 0.5);
    }
}

//------------------------------------  Set maximum delay
void
ChanSlice::setSliceTime(Interval Slice, Interval maxOff) {
    mSliceTime = Slice;
    if (Slice > Interval(0.0)) mNCkSlice = size_t(maxOff/Slice + 0.5);
    else                       mNCkSlice = 0;
}

static void
DumpIt(const short* data, int nw) {
    for (int i=0 ; i<nw ; i+=8) {
        printf("%5x  %7i %7i %7i %7i %7i %7i %7i %7i\n", i,
	       data[i  ], data[i+1], data[i+2], data[i+3], 
	       data[i+4], data[i+5], data[i+6], data[i+7]);
    }
}

static void
DumpIt(const float* data, int nw) {
    for (int i=0 ; i<nw ; i+=8) {
        printf("%5x  %7f %7f %7f %7f %7f %7f %7f %7f\n", i,
	       data[i  ], data[i+1], data[i+2], data[i+3], 
	       data[i+4], data[i+5], data[i+6], data[i+7]);
    }
}

//
//   Loop over slices and compare to successive data. The scan is performed
//   in 3 steps
//     1) Loop over slices in the saved data region. compare to new data.
//     2) Loop over all new data slices, compare to data other new slices.
//
int
ChanSlice::Scan(size_t nw, const short *data) {
    mNOverlap = 0;
    if (mSavType != kShort) freeData();

    //----------------------------------  Loop over saved data slices (inx)
    for (index_t inx=0 ; inx<mSaved ; inx+=mNWSlice) {
        const short* pi = reinterpret_cast<short*>(mSavData)+inx;

	//------------------------------  Make sure data aren't constant
	short first = *pi;
	bool constSlice(true);
	for (index_t k=1 ; k<mNWSlice ; k++) {
	    if (first != pi[k]) {
	        constSlice = false;
		break;
	    }
	}
	if (constSlice) continue;

        //------------------------------  Loop over Comparison slices (jnx)
        size_t maxComp = (mNCkSlice+1)*mNWSlice - mSaved + inx;
	if (nw < maxComp) maxComp = nw;
	for (index_t jnx=0 ; jnx<maxComp ; jnx+=mNWSlice) {
	    const short* pj = data+jnx;
	    if (*pi == *pj && !memcmp(pi, pj, sizeof(short)*mNWSlice)) {
	        if (mDebug) cout << "Match on channel: " << mChanName << endl;
		if (mDebug > 3) {
		    cout << "save[" << inx << "-" << inx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pi, mNWSlice);
		    cout << "data[" << jnx << "-" << jnx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pj, mNWSlice);
		}
		mNOverlap++;
	    }
	}
    }

    //----------------------------------  Loop over new data slices
    for (index_t inx=0 ; inx<nw ; inx+=mNWSlice) {
        const short* pi = data+inx;

	//------------------------------  Make sure data aren't constant
	short first = *pi;
	bool constSlice(true);
	for (index_t k=1 ; k<mNWSlice ; k++) {
	    if (first != pi[k]) {
	        constSlice = false;
		break;
	    }
	}
	if (constSlice) continue;

        //------------------------------  Loop over Comparison slices (jnx)
        size_t maxComp = inx + (mNCkSlice+1)*mNWSlice;
	if (maxComp > nw) maxComp = nw;
	for (index_t jnx=inx+mNWSlice ; jnx<maxComp ; jnx+=mNWSlice) {
	    const short* pj = data+jnx;
	    if (*pi == *pj && !memcmp(pi, pj, sizeof(short)*mNWSlice)) {
	        if (mDebug) cout << "Match on channel: " << mChanName << endl;
		if (mDebug > 3) {
		    cout << "data[" << inx << "-" << inx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pi, mNWSlice);
		    cout << "data[" << jnx << "-" << jnx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pj, mNWSlice);
		}
		mNOverlap++;
	    }
	}
    }

    //----------------------------------  Save incomplete slices
    size_t nSave = mNCkSlice*mNWSlice;
    saveData(data+nw-nSave, nSave);

    //----------------------------------  Return number of slice errors
    mStats.inc_test();
    if (mNOverlap) mStats.inc_sngl();
    return mNOverlap;
}

//======================================  Check everything
int
ChanSlice::Scan(size_t nw, const float *data) {
    mNOverlap = 0;
    if (mSavType != kFloat) freeData();

    //----------------------------------  Loop over saved data slices (inx)
    for (index_t inx=0 ; inx<mSaved ; inx+=mNWSlice) {
        const float* pi = reinterpret_cast<float*>(mSavData)+inx;

	//------------------------------  Make sure data aren't constant
	float first = *pi;
	bool constSlice(true);
	for (index_t k=1 ; k<mNWSlice ; k++) {
	    if (first != pi[k]) {
	        constSlice = false;
		break;
	    }
	}
	if (constSlice) continue;

        //------------------------------  Loop over Comparison slices (jnx)
        size_t maxComp = (mNCkSlice+1)*mNWSlice - mSaved + inx;
	if (nw < maxComp) maxComp = nw;
	for (index_t jnx=0 ; jnx<maxComp ; jnx+=mNWSlice) {
	    const float* pj = data+jnx;
	    if (*pi == *pj && !memcmp(pi, pj, sizeof(float)*mNWSlice)) {
	        if (mDebug) cout << "Match on channel: " << mChanName << endl;
		if (mDebug > 3) {
		    cout << "save[" << inx << "-" << inx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pi, mNWSlice);
		    cout << "data[" << jnx << "-" << jnx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pj, mNWSlice);
		}
		mNOverlap++;
	    }
	}
    }

    //----------------------------------  Loop over new data slices
    for (index_t inx=0 ; inx<nw ; inx+=mNWSlice) {
        const float* pi = data+inx;

	//------------------------------  Make sure data aren't constant
	float first = *pi;
	bool constSlice(true);
	for (index_t k=1 ; k<mNWSlice ; k++) {
	    if (first != pi[k]) {
	        constSlice = false;
		break;
	    }
	}
	if (constSlice) continue;

        //------------------------------  Loop over Comparison slices (jnx)
        size_t maxComp = inx + (mNCkSlice+1)*mNWSlice;
	if (maxComp > nw) maxComp = nw;
	for (index_t jnx=inx+mNWSlice ; jnx<maxComp ; jnx+=mNWSlice) {
	    const float* pj = data+jnx;
	    if (*pi == *pj && !memcmp(pi, pj, sizeof(float)*mNWSlice)) {
	        if (mDebug) cout << "Match on channel: " << mChanName << endl;
		if (mDebug > 3) {
		    cout << "data[" << inx << "-" << inx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pi, mNWSlice);
		    cout << "data[" << jnx << "-" << jnx+mNWSlice << "]" 
			 << endl;
		    DumpIt(pj, mNWSlice);
		}
		mNOverlap++;
	    }
	}
    }

    //----------------------------------  Save incomplete slices
    size_t nSave = mNCkSlice*mNWSlice;
    saveData(data+nw-nSave, nSave);

    //----------------------------------  Return number of slice errors
    mStats.inc_test();
    if (mNOverlap) mStats.inc_sngl();
    return mNOverlap;
}
//======================================  Check everything
int
ChanSlice::Scan(size_t nw, const fComplex *data) {
    mNOverlap = 0;
    if (mSavType != kCmplx) freeData();

    //----------------------------------  Loop over saved data slices (inx)
    for (index_t inx=0 ; inx<mSaved ; inx+=mNWSlice) {
        const fComplex* pi = reinterpret_cast<fComplex*>(mSavData)+inx;

	//------------------------------  Make sure data aren't constant
	fComplex first = *pi;
	bool constSlice(true);
	for (index_t k=1 ; k<mNWSlice ; k++) {
	    if (first != pi[k]) {
	        constSlice = false;
		break;
	    }
	}
	if (constSlice) continue;

        //------------------------------  Loop over Comparison slices (jnx)
        size_t maxComp = (mNCkSlice+1)*mNWSlice - mSaved + inx;
	if (nw < maxComp) maxComp = nw;
	for (index_t jnx=0 ; jnx<maxComp ; jnx+=mNWSlice) {
	    const fComplex* pj = data+jnx;
	    if (*pi == *pj && !memcmp(pi, pj, sizeof(fComplex)*mNWSlice)) {
	        if (mDebug) cout << "Match on channel: " << mChanName << endl;
		if (mDebug > 3) {
		    cout << "save[" << inx << "-" << inx+mNWSlice << "]" 
			 << endl;
		    DumpIt(reinterpret_cast<const float*>(pi), 2*mNWSlice);
		    cout << "data[" << jnx << "-" << jnx+mNWSlice << "]" 
			 << endl;
		    DumpIt(reinterpret_cast<const float*>(pj), 2*mNWSlice);
		}
		mNOverlap++;
	    }
	}
    }

    //----------------------------------  Loop over new data slices
    for (index_t inx=0 ; inx<nw ; inx+=mNWSlice) {
        const fComplex* pi = data+inx;

	//------------------------------  Make sure data aren't constant
	fComplex first = *pi;
	bool constSlice(true);
	for (index_t k=1 ; k<mNWSlice ; k++) {
	    if (first != pi[k]) {
	        constSlice = false;
		break;
	    }
	}
	if (constSlice) continue;

        //------------------------------  Loop over Comparison slices (jnx)
        size_t maxComp = inx + (mNCkSlice+1)*mNWSlice;
	if (maxComp > nw) maxComp = nw;
	for (index_t jnx=inx+mNWSlice ; jnx<maxComp ; jnx+=mNWSlice) {
	    const fComplex* pj = data+jnx;
	    if (*pi == *pj && !memcmp(pi, pj, sizeof(fComplex)*mNWSlice)) {
	        if (mDebug) cout << "Match on channel: " << mChanName << endl;
		if (mDebug > 3) {
		    cout << "data[" << inx << "-" << inx+mNWSlice << "]" 
			 << endl;
		    DumpIt(reinterpret_cast<const float*>(pi), 2*mNWSlice);
		    cout << "data[" << jnx << "-" << jnx+mNWSlice << "]" 
			 << endl;
		    DumpIt(reinterpret_cast<const float*>(pj), 2*mNWSlice);
		}
		mNOverlap++;
	    }
	}
    }

    //----------------------------------  Save incomplete slices
    size_t nSave = mNCkSlice*mNWSlice;
    saveData(data+nw-nSave, nSave);

    //----------------------------------  Return number of slice errors
    mStats.inc_test();
    if (mNOverlap) mStats.inc_sngl();
    return mNOverlap;
}

//======================================  Save data
void 
ChanSlice::saveData(const short* p, size_t N) {
    if (mSavType != kShort || mSaved < N) {
        freeData();
        if (mDebug) cout << "Buy " << N << " words for " 
			 << mChanName << endl;
	mSavData = new short[N];
	mSavType = kShort;
    }
    memcpy(mSavData, p, N*sizeof(short));
    mSaved = N;
}

void 
ChanSlice::saveData(const float* p, size_t N) {
    if (mSavType != kFloat || mSaved < N) {
        freeData();
        if (mDebug) cout << "Buy " << N << " words for " 
			 << mChanName << endl;
	mSavData = new float[N];
	mSavType = kFloat;
    }
    memcpy(mSavData, p, N*sizeof(float));
    mSaved = N;
}

void 
ChanSlice::saveData(const fComplex* p, size_t N) {
    if (mSavType != kCmplx || mSaved < N) {
        freeData();
        if (mDebug) cout << "Buy " << N << " words for " 
			 << mChanName << endl;
	mSavData = new fComplex[N];
	mSavType = kCmplx;
    }
    memcpy(mSavData, p, N*sizeof(fComplex));
    mSaved = N;
}

//======================================  Free data
void 
ChanSlice::freeData(void) {
    switch (mSavType) {
    case kShort:
      delete[] reinterpret_cast<short*>(mSavData);
      break;
    case kFloat:
      delete[] reinterpret_cast<float*>(mSavData);
      break;
    case kCmplx:
      delete[] reinterpret_cast<fComplex*>(mSavData);
      break;
    default:
      break;
    }
    mSavType = kNone;
    mSaved   = 0;
}

//======================================  Check everything
void
ChanSlice::resetStats(void) {
    mTotal += mStats;
    mStats.clear();
}
